"""
union proc
"""
from collections import OrderedDict

from flume import node


class union(node):

    name = 'union'

    def __init__(self, *fieldnames):
        node.__init__(self)
        self.fieldnames = fieldnames

    def loop(self):
        points = []
        previous_point = None

        if len(self.fieldnames) == 0:
            while self.running:
                points = self.pull()

                for point in points:
                    if previous_point is None:
                        previous_point = point

                    elif previous_point.time == point.time:
                        previous_point.update(point)

                    else:
                        self.push(previous_point)
                        previous_point = point
            
            if previous_point is not None:
                self.push(previous_point)

        else:
            # when unioning on non `time` field we basically buffer every
            # point and emit when we have a complete window

            # XXX: non time based union should have a limit on maximum
            #      amount of points to handle like 100K so we don't run out of
            #      memory
            result = OrderedDict()
            fieldnames = self.fieldnames

            while self.running:
                points = self.pull()

                for point in points:
                    key = '.'.join([str(point[fname]) for fname in fieldnames])

                    if key not in result:
                        result[key] = point

                    else:
                        result[key].update(point)

            self.push([point for point in result.values()])



"""
flume moment module

- used to simplify the representation of dates and durations by easily
  parseable human expressions
"""

from datetime import datetime, timedelta
from time import mktime

import parsedatetime
import pytz

from delta import parse
from dateutil import parser

from flume.exceptions import FlumineException
from flume.util import is_string


def __make_utc(dt):
    """
    localize or attach the UTC timezone to the datetime object provided
    """
    return pytz.UTC.localize(dt)

__MIN = __make_utc(datetime(1900, 1, 1))
__MAX = __make_utc(datetime.max)


def now():
    """
    returns a datetime object that represents the current time
    """
    return __make_utc(datetime.now())


def beginning():
    """
    returns a datetime object that represents the beginning of time
    """
    return __MIN


def end():
    """
    returns a datetime object that represents the end of time
    """
    return __MAX


def forever():
    """
    returns a timedelta that represents forever """
    return timedelta.max


def date(string):
    """
    returns the datetime associated with the date expressed in the string
    provided. You can express time with an exactly ISO8601 date or you can
    also express it using relative time expression such as '1 hour ago'
    """
    if isinstance(string, datetime):
        return string

    elif is_string(string):
        dt = None
        try:
            dt = parser.parse(string)
        except ValueError as exception:
            if str(exception) == 'Unknown string format':
                # lets parse the possible relative time with parsedatetime
                struct_time, result = parsedatetime.Calendar().parse(string)

                if result == 0:
                    raise FlumineException('unable to parse date "%s"' % string)

                dt = datetime.fromtimestamp(mktime(struct_time))

        if dt is None:
            raise FlumineException('Unable to parse moment "%s"' % string)

        if dt.tzinfo is None:
            # all dates without a timezone are assumed to be in UTC
            dt = __make_utc(dt)

        return dt
    else:
        raise FlumineException('Unable to parse moment "%s"' % string)

def duration(string, context=None):
    """
    Converts a string representing a duration into a timedelta

    arguments:

        string - The string to convert into seconds representing the duration
                 expressed in natural human language, ie 1 hour, 1h, 2 days,
                 etc.

        context - Datetime object expressing the context to evaluate the
                  duration string in. This really applies to things like
                  `1 month` in the context of January has 31 days vs April has
                  30 days. Defaults to datetime.now()
    returns:

        Number of seconds represented by the duration string provided
    """
    if isinstance(string, timedelta):
        return string

    elif is_string(string):
        delta = parse(string, context=context)

        if delta is None:
            raise FlumineException('unable to parse the duration "%s"' % string)

        return delta

    raise FlumineException('unable to parse the duration "%s"' % string)

def datetime_to_iso8601(dt):
    """
    Convert a datetime into the ISO8601 string that it represents
    """
    return '%s.%03dZ' % (dt.strftime('%Y-%m-%dT%H:%M:%S'),
                         int(dt.microsecond / 1000))



"""
emit source unittests
"""
import unittest

from robber import expect

from flume import *


class EmitTest(unittest.TestCase):

    def test_emit_no_points(self):
        results = []
        (
            emit(points=[])
            | memory(results)
        ).execute()
        expect(results).to.eq([])

    def test_emit_some_points(self):
        results = []
        data = [{'value': 1}, {'value': 2}, {'value': 3}]
        (
            emit(points=data)
            | memory(results)
        ).execute()
        expect(results).to.eq(data)

    def test_emit_with_limit(self):
        results = []
        (
            emit(limit=10, every='0.01s')
            | memory(results)
        ).execute()
        expect(results).to.have.length(10)

    def test_emit_starting_at_specific_date(self):
        results = []
        (
            emit(start='2015-01-01', limit=5, every='1s')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'time': '2015-01-01T00:00:00.000Z'},
            {'time': '2015-01-01T00:00:01.000Z'},
            {'time': '2015-01-01T00:00:02.000Z'},
            {'time': '2015-01-01T00:00:03.000Z'},
            {'time': '2015-01-01T00:00:04.000Z'}
        ])

    def test_emit_with_every_month_expression(self):
        results = []
        (
            emit(start='2015-01-01', limit=5, every='1M')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'time': '2015-01-01T00:00:00.000Z'},
            {'time': '2015-02-01T00:00:00.000Z'},
            {'time': '2015-03-01T00:00:00.000Z'},
            {'time': '2015-04-01T00:00:00.000Z'},
            {'time': '2015-05-01T00:00:00.000Z'}
        ])



"""
elastic adapter core module
"""

import json

from elasticsearch import Elasticsearch, helpers

from flume import logger, moment
from flume.adapters.adapter import adapter
from flume.adapters.elastic.query import filter_to_es_query
from flume.exceptions import FlumineException
from flume.point import Point


class elastic(adapter):
    """
    elastic flume adapter
    """

    name = 'elastic'

    def __init__(self,
                 index='_all',
                 type='metric',
                 host='localhost',
                 port=9200,
                 filter=None,
                 batch=1024):
        self.index = index
        self.type = type
        self.host = host
        self.port = port
        self.filter = filter
        self.batch = batch
        self.clients = {}

    def _get_elasticsearch(self, host, port):
        """
        cache ES client
        """
        key = '%s:%s' % (host, port)

        if key not in self.clients:
            self.clients[key] = Elasticsearch([host], port=port)

        return self.clients[key]

    def read(self):
        """
        read points out of ES
        """
        points = []
        client = self._get_elasticsearch(self.host, self.port)

        query = filter_to_es_query(self.filter)
        logger.debug('es query %s' % json.dumps(query))

        for result in helpers.scan(client,
                                   index=self.index,
                                   query={'query': query, 'sort': ['time']},
                                   preserve_order=True):
            point = Point(**result['_source'])
            points.append(point)

            if len(points) > self.batch:
                yield points
                points = []

        if len(points) > 0:
            yield points

    def write(self, points):
        """
        write data points to the index in the ES instance indicated
        """
        client = self._get_elasticsearch(self.host, self.port)

        pushing = []

        for point in points:
            point = point.json()
            point['_index'] = self.index
            point['_type'] = self.type
            pushing.append(point)

        _, errors = helpers.bulk(client, pushing)

        for error in errors:
            logger.error(error)
            raise FlumineException('errors while writing to elasticsearch')

    def eof(self):
        pass



import unittest

from robber import expect
from flume import *
from flume.exceptions import FlumineException

from flume.core import register_sink, register_proc, register_source

class SyslogTest(unittest.TestCase):
    """
    verify that the third-party utilities for syslog work as expected
    """

    def test_syslog_fails_when_path_does_not_exist(self):
        with self.assertRaisesRegexp(OSError, 'No such file or directory: \'/no/mans/land/\''):
            syslog(path='/no/mans/land/').execute()

    def test_syslog_can_find_a_single_syslog_by_path(self):
        results = []

        (
            syslog(path='test/unit/thirdparty/')
            | reduce(count=count())
            | keep('count')
            | memory(results)
        ).execute()

        expect(results).to.eq([
            {'count':  274}
        ])

    def test_syslog_read_syslog_file_directly(self):
        results = []

        (
            syslog(logfile='test/unit/thirdparty/syslog')
            | reduce(count=count())
            | keep('count')
            | memory(results)
        ).execute()

        expect(results).to.eq([
            {'count':  274}
        ])

    def test_syslog_can_read_compressed_and_not_compressed_syslog_files(self):
        results = []

        (
            syslog(path='test/unit/thirdparty/logs/')
            | reduce(count=count())
            | keep('count')
            | memory(results)
        ).execute()

        expect(results).to.eq([
            {'count':  1200}
        ])

    def test_syslog_can_handle_compressed_direct_logfile(self):
        results = []

        (
            syslog(logfile='test/unit/thirdparty/logs/syslog.1.gz')
            | reduce(count=count())
            | keep('count')
            | memory(results)
        ).execute()

        expect(results).to.eq([
            {'count':  926}
        ])



"""
date reducer - exposes the date manipulation functions as reducers
"""
import inspect

from flume import reducer


def strftime(fieldname, fmt):

    class strftime_reducer(reducer):

        def __init__(self, fieldname, fmt):
            self.fieldname = fieldname
            self.fmt = fmt
            self.value = None

        def update(self, point):
            self.value = point[self.fieldname].strftime(self.fmt)

        def result(self):
            return self.value

        def reset(self):
            self.value = None

    return strftime_reducer(fieldname, fmt)

class date(object):

    @classmethod
    def year(cls, fieldname):
        """
        return the year for the field specified. example: 2010, 2016, etc.
        """
        return strftime(fieldname, '%Y')

    @classmethod
    def fullmonth(cls, fieldname):
        """
        return the full month name for the field specified. example: January
        """
        return strftime(fieldname, '%B')

    @classmethod
    def month(cls, fieldname):
        """
        return the abbreviated month name for the field specified. example: Jan
        """
        return strftime(fieldname, '%b')

    @classmethod
    def fullweekday(cls, fieldname):
        """
        return the full weekday name for the field specified. example: Sunday
        """
        return strftime(fieldname, '%A')

    @classmethod
    def weekday(cls, fieldname):
        """
        return the abbreviated weekday name for the field specified. example: Sun
        """
        return strftime(fieldname, '%a')

    @classmethod
    def strftime(cls, fieldname, fmt):
        """
        return the strftime format(fmt) applied to the field name provided
        (must be a datetime object) format details: https://goo.gl/r8rDLB
        """
        return strftime(fieldname, fmt)



"""
keep processor
"""
from flume import node, Point


class keep(node):

    name = 'keep'

    def __init__(self, *fieldnames):
        node.__init__(self, *fieldnames)
        self.fieldnames = fieldnames

    def loop(self):
        while self.running:
            points = self.pull()
            points_out = []

            for point in points:
                new_point = Point()
                points_out.append(new_point)

                for fieldname in self.fieldnames:
                    if point.hasfield(fieldname):
                        new_point[fieldname] = point.lookup(fieldname)

            self.push(points_out)



"""
flume logger
"""

import logging

from logging import DEBUG, INFO, ERROR, WARN, CRITICAL


def init():
    """
    internal method that initialize the flume logger
    """
    logger = logging.getLogger('flume')
    streamHandler = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
    streamHandler.setFormatter(formatter)
    logger.addHandler(streamHandler)
    logger.setLevel(logging.WARN)


def setLogLevel(lvl):
    """
    set logging level (ie logger.INFO, DEBUG, etc.)
    """
    logging.getLogger('flume').setLevel(lvl)


def debug(*args, **kwargs):
    """
    log a debug message
    """
    logging.getLogger('flume').debug(*args, **kwargs)


def info(*args, **kwargs):
    """
    log an info message
    """
    logging.getLogger('flume').info(*args, **kwargs)


def warn(*args, **kwargs):
    """
    log a warning message
    """
    logging.getLogger('flume').warn(*args, **kwargs)


def error(*args, **kwargs):
    """
    log an error message
    """
    logging.getLogger('flume').error(*args, **kwargs)


def critical(*args, **kwargs):
    """
    log a critical message
    """
    logging.getLogger('flume').critical(*args, **kwargs)


init()



"""
pygal view
"""

import sys
import tempfile

import pygal

from flume.sinks.views.base import base
from flume.exceptions import FlumineException
from flume import moment

HTML_TEMPLATE = """
<!DOCTYPE html>
<html>
  <head>
  <script type="text/javascript" src="http://kozea.github.com/pygal.js/javascripts/svg.jquery.js"></script>
  <script type="text/javascript" src="http://kozea.github.com/pygal.js/javascripts/pygal-tooltips.js"></script>
    <!-- ... -->
  </head>
  <body>
    <figure>
        %s
    </figure>
  </body>
</html>"""


def is_ipython():
    try:
        return __IPYTHON__
    except:
        return False

class PyGal(base):

    def __init__(self,
                 title='',
                 width=1280,
                 height=1024,
                 filename=None,
                 format='png'):
        self.title = title
        self.width = width
        self.height = height

        if not is_ipython():
            if format not in ['png', 'html']:
                raise FlumineException('supported formats are "html" and "png"')

            if filename is None:
                # save to a temporary file and open with default image viewer
                _, filename = tempfile.mkstemp('.' + format)
                sys.stdout.write('rendering chart to "%s"\n' % filename)
                sys.stdout.flush()

            self.filename = filename

    def render(self, chart_type, data):
        if chart_type == 'linechart':
            chart = pygal.XY(height=self.height,
                             width=self.width,
                             x_label_rotation=45)

            chart.title = self.title

            for series_name in data.keys():
                series = data[series_name]
                chart.add(series_name, series)

        elif chart_type == 'timechart':
            chart = pygal.DateTimeLine(height=self.height,
                                       width=self.width,
                                       x_label_rotation=45,
                                       x_value_formatter=moment.datetime_to_iso8601)

            chart.title = self.title

            for series_name in data.keys():
                series = data[series_name]
                chart.add(series_name, series)

        elif chart_type == 'barchart':
            chart = pygal.Bar(height=self.height, width=self.width)
            chart.title = self.title

            chart.x_labels = [category for (category, _) in data[list(data.keys())[0]]]

            for series_name in data.keys():
                series = [value for (_, value) in data[series_name]]
                chart.add(series_name, series)

        else:
            raise FlumineException('unsupported chart type "%s" for pygal view'
                                   % chart_type)

        if is_ipython():
            from IPython.core.display import display, HTML, clear_output
            clear_output(wait=True)
            display(HTML(chart.render()))

        elif self.filename.endswith('.png'):
            chart.render_to_png(self.filename)

        elif self.filename.endswith('.html'):
            with open(self.filename, 'w') as output:
                output.write(HTML_TEMPLATE % chart.render())



"""
reducer proc
"""

import copy

from flume import moment, node, Point, reducer
from flume import util


class reduce(node):

    name = 'reduce'

    def __init__(self,
                 every=moment.forever(),
                 reset=True,
                 by=None,
                 **fields):

        node.__init__(self,
                      every=every,
                      reset=reset,
                      **fields)

        self.every = every
        self.reset = reset

        if by is None:
            by = []

        self.by = by
        self.fields = fields

    def loop(self):
        results = {}
        fields = {}
        batch_start = None
        batch_end = None

        def init_results(results, fields, by_key, init_point, point):
            results[by_key] = init_point

            for by in self.by:
                results[by_key].set(by, point[by])

            if by_key not in fields:
                fields[by_key] = {}
                for (field, value) in self.fields.items():
                    fields[by_key][field] = copy.deepcopy(value)

        while self.running:
            points = self.pull()

            for point in points:
                by_key = ':'.join([util.u(point[by]) for by in self.by])

                if 'time' in point:
                    if batch_start is None:
                        batch_start = point.time

                        if self.every == moment.forever():
                            batch_end = moment.end()

                        else:
                            every = moment.duration(self.every, context=batch_start)
                            batch_end = batch_start + every

                        init_results(results, fields, by_key, Point(time=batch_start), point)

                    if point.time >= batch_end:
                        # lets make sure that we've had a single batch
                        # of data received and not multiple empty batches
                        gaps = int((point.time - batch_start).total_seconds()/every.total_seconds())

                        # if we reached the end of a batch then push
                        self.push(results.values())

                        # new batch
                        batch_start = batch_end
                        every = moment.duration(self.every, context=batch_start)
                        batch_end = batch_end + every

                        # XXX: to carry over any fields that were found in a
                        # batch but not found in a subsequent batch we should
                        # copy the existing results here to the new one

                        results = {}
                        init_results(results, fields, by_key, Point(time=batch_start), point)

                        # reset reducers
                        if self.reset:
                            for by_field_key in fields.keys():
                                for (_, value) in fields[by_field_key].items():
                                    if isinstance(value, reducer):
                                        value.reset()

                        if gaps > 1:
                            # generate empty batches since the last batch_start
                            for _ in range(1, gaps):
                                empty_results = {}
                                empty_fields = {}

                                for by_key in results:
                                    init_results(empty_results,
                                                 empty_fields,
                                                 by_key,
                                                 Point(time=batch_start),
                                                 point)

                                    fields_copy = fields[by_key].copy()
                                    for (key, value) in fields_copy.items():
                                        if isinstance(value, reducer):
                                            fields_copy[key] = value.result()

                                        else:
                                            fields_copy[key] = value

                                        empty_results[by_key][key] = fields_copy[key]

                                batch_start = batch_end
                                every = moment.duration(self.every, context=batch_start)
                                batch_end = batch_end + every

                                # if we reached the end of a batch then push
                                self.push(empty_results.values())

                            results = {}
                            init_results(results, fields, by_key, Point(time=batch_start), point)

                    if by_key not in results:
                        init_results(results, fields, by_key, Point(time=batch_start), point)

                else:
                    if by_key not in results:
                        init_results(results, fields, by_key, Point(), point)

                # update them reducers
                fields_copy = fields[by_key].copy()
                for (key, value) in fields_copy.items():
                    if isinstance(value, reducer):
                        value.update(point)
                        fields_copy[key] = value.result()

                    else:
                        if util.is_string(value):
                            fields_copy[key] = value.format(**point.json())

                        else:
                            fields_copy[key] = value

                    results[by_key][key] = fields_copy[key]

        self.push(results.values())



"""
flume utilities
"""

import sys

IS_PY2 = sys.version[0] == '2'

def is_string(string):
    """
    check an object is a string across python 2 and 3
    """
    if IS_PY2:
        return isinstance(string, basestring)
    else:
        return isinstance(string, str)

def u(string):
    """
    return a string as a unicode string in python 2 and 3
    """
    if IS_PY2:
        return unicode(string)
    else:
        return str(string)



"""
gnuplot view
"""
import os
import subprocess
import tempfile

from datetime import datetime

import blessings

from flume.sinks.views.base import base
from flume.exceptions import FlumineException


class Gnuplot(base):

    def __init__(self,
                 title='',
                 width=1280,
                 height=1024,
                 terminal='dumb',
                 timefmt='%Y/%d/%m-%H:%M:%S'):
        self.title = title
        self.width = width
        self.height = height
        self.terminal = terminal
        self.timefmt = timefmt

        if terminal == 'dumb':
            self.process = None

        else:
            self.process = subprocess.Popen(['gnuplot', '-p'], stdin=subprocess.PIPE)

    def init_terminal(self, min_value, max_value):
        if self.terminal == 'dumb':
            if self.process is not None:
                self.write('exit\n')
                self.process.wait()

            import time
            time.sleep(0.1)
            self.process = subprocess.Popen(['gnuplot'], stdin=subprocess.PIPE)

            terminal = blessings.Terminal()

            self.write('set terminal %s size %d, %d\n' %
                       (self.terminal, terminal.width, terminal.height))

        else:
            self.write('clear\n')
            self.write('set terminal %s size %d, %d\n' %
                       (self.terminal, self.width, self.height))

        max_value = max_value * 1.25
        self.write('set yrange [%g:%g]\n' % (min_value, max_value))

    def write(self, line):
        self.process.stdin.write(line.encode('utf-8'))
        self.process.stdin.flush()

    def render(self, chart_type, data):
        min_value = 0
        max_value = 0
        x_is_time = False

        tmp = tempfile.mkdtemp()
        for series_name in data.keys():
            filename = os.path.join(tmp, '%s.dat' % series_name)

            with open(filename, 'w') as output:
                for (colx, coly) in data[series_name]:

                    if coly > max_value or max_value is None:
                        max_value = coly

                    if coly < min_value or min_value is None:
                        min_value = coly

                    if isinstance(colx, datetime):
                        x_is_time = True
                        date = '%s/%s/%s-%s:%s:%s' % (colx.year, colx.day, colx.month,
                                                      colx.hour, colx.minute, colx.second)
                        output.write('%s\t%s\n' % (date, coly))

                    else:
                        output.write('%s\t%s\n' % (colx, coly))

        self.init_terminal(min_value, max_value)

        if chart_type == 'linechart' or chart_type == 'timechart':
            if x_is_time:

                if chart_type == 'timechart':
                    self.write('set xdata time\n')
                    self.write('set format x "%s"\n' % self.timefmt)
                    self.write('set timefmt "%s"\n' % self.timefmt)

            series_string = ['"%s" using 1:2 with linespoints title "%s"' % \
                             (os.path.join(tmp, '%s.dat' % series_name), series_name)
                             for series_name in data.keys()]

        elif chart_type == 'barchart':
            self.write('set boxwidth 0.8\n')
            self.write('set style fill solid\n')
            self.write('set style data histogram\n')
            self.write('set style histogram cluster gap 1\n')

            series_string = ['"%s" using 2:xtic(1) title "%s"' % \
                             (os.path.join(tmp, '%s.dat' % series_name), series_name)
                             for series_name in data.keys()]

        else:
            raise FlumineException('unsupported chart type "%s" for pygal view'
                                   % chart_type)

        self.write('plot %s\n' % ', '.join(series_string))



"""
reduce proc unittests
"""
import unittest

from robber import expect
from flume import *


class ReduceTest(unittest.TestCase):

    def test_reduce_to_a_static_value(self):
        results = []
        (
            emit(limit=5, every='0.001s')
            | reduce(count=0)
            | keep('count')
            | memory(results)
        ).execute()
        expect(results).to.eq([{'count': 0}])

    def test_reduce_field_with_interpolated_field_value(self):
        results = []
        (
            emit(limit=5, every='0.001s')
            | put(count=count())
            | reduce(message='count is {count}')
            | keep('message')
            | memory(results)
        ).execute()
        expect(results).to.eq([{'message': 'count is 5'}])

    def test_reduce_field_with_interpolated_time_value(self):
        results = []
        (
            emit(limit=5, start='2016-01-01')
            | reduce(message='last point at "{time}"')
            | keep('message')
            | memory(results)
        ).execute()
        expect(results).to.eq([{'message': 'last point at "2016-01-01T00:00:04.000Z"'}])

    def test_reduce_to_a_count(self):
        results = []
        (
            emit(limit=5, every='0.001s')
            | reduce(count=count())
            | keep('count')
            | memory(results)
        ).execute()
        expect(results).to.eq([{'count': 5}])

    def test_reduce_to_the_maximum(self):
        results = []
        (
            emit(limit=5, every='0.001s')
            | put(count=count())
            | reduce(maximum=maximum('count'))
            | keep('maximum')
            | memory(results)
        ).execute()
        expect(results).to.eq([{'maximum': 5}])

    def test_reduce_to_the_minimum(self):
        results = []
        (
            emit(limit=5, every='0.001s')
            | put(count=count())
            | reduce(minimum=minimum('count'))
            | keep('minimum')
            | memory(results)
        ).execute()

        expect(results).to.eq([{'minimum': 1}])

    def test_reduce_with_custom_reducer(self):

        class count_odds(reducer):

            def __init__(self, fieldname):
                self.count = 0
                self.fieldname = fieldname

            def update(self, point):
                if self.fieldname in point:
                    if point[self.fieldname] % 2 != 0:
                        self.count += 1

            def result(self):
                return self.count

            def reset(self):
                self.count = 0

        results = []
        (
            emit(limit=10, every='0.001s')
            | put(count=count())
            | reduce(odds=count_odds('count'))
            | keep('odds')
            | memory(results)
        ).execute()

        expect(results).to.eq([{'odds': 5}])

    def test_reduce_with_mixed_fields(self):
        results = []
        (
            emit(limit=10, every='0.001s')
            | reduce(count=count(), foo='bar', someid=1)
            | keep('count', 'foo', 'someid')
            | memory(results)
        ).execute()

        expect(results).to.eq([
            {'count': 10, 'foo': 'bar', 'someid': 1}
        ])

    def test_reduce_to_a_count_by_a_field(self):
        results = []
        (
            emit(limit=10, every='0.001s')
            | (
                put(foo='bar'),
                put(foo='baz')
            ) | reduce(count=count(), by=['foo'])
            | keep('foo', 'count')
            | memory(results)
        ).execute()

        expect(results).to.have.length(2)
        expect(results).to.contain({'foo': 'bar', 'count': 10})
        expect(results).to.contain({'foo': 'baz', 'count': 10})

    def test_reduce_to_a_maximum_by_a_field(self):
        results = []
        (
            emit(limit=10, every='0.001s')
            | put(count=count())
            | (
                put(foo='bar'),
                put(foo='baz')
            ) | reduce(maximum=maximum('count'), by=['foo'])
            | keep('foo', 'maximum')
            | memory(results)
        ).execute()
        expect(results).to.have.length(2)
        expect(results).to.contain({'foo': 'bar', 'maximum': 10})
        expect(results).to.contain({'foo': 'baz', 'maximum': 10})

    def test_reduce_to_a_minimum_by_a_field(self):
        results = []
        (
            emit(limit=10, every='0.001s')
            | put(count=count())
            | (
                put(foo='bar'),
                put(foo='baz')
            ) | reduce(minimum=minimum('count'), by=['foo'])
            | keep('foo', 'minimum')
            | memory(results)
        ).execute()
        expect(results).to.have.length(2)
        expect(results).to.contain({'foo': 'bar', 'minimum': 1})
        expect(results).to.contain({'foo': 'baz', 'minimum': 1})

    def test_reduce_to_a_count_by_multiple_fields(self):
        results = []
        (
            emit(limit=10, every='0.001s')
            | (
                put(a=iterate([2]), b=iterate([2])),
                put(a=iterate([2, 3]), b=iterate([1, 2]))
            ) | reduce(count=count(), by=['a', 'b'])
            | keep('a', 'b', 'count')
            | memory(results)
        ).execute()

        expect(results).to.have.length(3)
        expect(results).to.contain({'a': 2, 'b': 2, 'count': 10})
        expect(results).to.contain({'a': 2, 'b': 1, 'count': 5})
        expect(results).to.contain({'a': 3, 'b': 2, 'count': 5})

    def test_reduce_with_mixed_fields_by_a_field(self):
        results = []
        (
            emit(limit=10, every='0.001s')
            | (
                put(foo='bar'),
                put(foo='baz')
            ) | reduce(count=count(), fizz='buzz', someid=1, by=['foo'])
            | keep('fizz', 'someid', 'foo', 'count')
            | memory(results)
        ).execute()

        expect(results).to.have.length(2)
        expect(results).to.contain({
            'fizz': 'buzz',
            'someid': 1,
            'foo': 'bar',
            'count': 10
        })

        expect(results).to.contain({
            'fizz': 'buzz',
            'someid': 1,
            'foo': 'baz',
            'count': 10
        })

    def test_reduce_every_count(self):
        results = []
        (
            emit(limit=10, every='1s', start='2016-01-01')
            | reduce(count=count(), every='3s')
            | memory(results)
        ).execute()

        expect(results).to.eq([
            {'time': '2016-01-01T00:00:00.000Z', 'count': 3},
            {'time': '2016-01-01T00:00:03.000Z', 'count': 3},
            {'time': '2016-01-01T00:00:06.000Z', 'count': 3},
            {'time': '2016-01-01T00:00:09.000Z', 'count': 1}
        ])

    def test_reduce_every_maximum(self):
        results = []
        (
            emit(limit=10, every='1s', start='2016-01-01')
            | put(count=count())
            | reduce(maximum=maximum('count'), every='3s')
            | memory(results)
        ).execute()

        expect(results).to.eq([
            {'time': '2016-01-01T00:00:00.000Z', 'maximum': 3},
            {'time': '2016-01-01T00:00:03.000Z', 'maximum': 6},
            {'time': '2016-01-01T00:00:06.000Z', 'maximum': 9},
            {'time': '2016-01-01T00:00:09.000Z', 'maximum': 10}
        ])

    def test_reduce_every_minimum(self):
        results = []
        (
            emit(limit=10, every='1s', start='2016-01-01')
            | put(count=count())
            | reduce(minimum=minimum('count'), every='3s')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'time': '2016-01-01T00:00:00.000Z', 'minimum': 1},
            {'time': '2016-01-01T00:00:03.000Z', 'minimum': 4},
            {'time': '2016-01-01T00:00:06.000Z', 'minimum': 7},
            {'time': '2016-01-01T00:00:09.000Z', 'minimum': 10}
        ])

    def test_reduce_every_count_with_mixed_fields(self):
        results = []
        (
            emit(limit=10, every='1s', start='2016-01-01')
            | reduce(count=count(), foo='bar', someid=1, every='3s')
            | memory(results)
        ).execute()

        expect(results).to.eq([
            {
                'time': '2016-01-01T00:00:00.000Z',
                'count': 3,
                'foo': 'bar',
                'someid': 1
            },
            {
                'time': '2016-01-01T00:00:03.000Z',
                'count': 3,
                'foo': 'bar',
                'someid': 1
            },
            {
                'time': '2016-01-01T00:00:06.000Z',
                'count': 3,
                'foo': 'bar',
                'someid': 1
            },
            {
                'time': '2016-01-01T00:00:09.000Z',
                'count': 1,
                'foo': 'bar',
                'someid': 1
            }
        ])

    def test_reduce_every_count_by_a_field(self):
        results = []
        (
            emit(limit=10, start='2016-01-01', every='1s')
            | (
                put(foo='bar'),
                put(foo='baz')
            ) | reduce(count=count(), every='2s', by=['foo'])
            | memory(results)
        ).execute()
        expect(results).to.have.length(10)

        for point in [
                {'time': '2016-01-01T00:00:00.000Z', 'foo': 'bar', 'count': 2},
                {'time': '2016-01-01T00:00:00.000Z', 'foo': 'baz', 'count': 2},
                {'time': '2016-01-01T00:00:02.000Z', 'foo': 'bar', 'count': 2},
                {'time': '2016-01-01T00:00:02.000Z', 'foo': 'baz', 'count': 2},
                {'time': '2016-01-01T00:00:04.000Z', 'foo': 'bar', 'count': 2},
                {'time': '2016-01-01T00:00:04.000Z', 'foo': 'baz', 'count': 2},
                {'time': '2016-01-01T00:00:06.000Z', 'foo': 'bar', 'count': 2},
                {'time': '2016-01-01T00:00:06.000Z', 'foo': 'baz', 'count': 2},
                {'time': '2016-01-01T00:00:08.000Z', 'foo': 'bar', 'count': 2},
                {'time': '2016-01-01T00:00:08.000Z', 'foo': 'baz', 'count': 2}]:
            expect(results).to.contain(point)

    def test_reduce_every_with_no_reset(self):
        results = []
        (
            emit(limit=10, every='1s', start='2016-01-01')
            | reduce(count=count(), every='3s', reset=False)
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'time': '2016-01-01T00:00:00.000Z', 'count': 3},
            {'time': '2016-01-01T00:00:03.000Z', 'count': 6},
            {'time': '2016-01-01T00:00:06.000Z', 'count': 9},
            {'time': '2016-01-01T00:00:09.000Z', 'count': 10}
        ])

    def test_reduce_every_with_no_reset_by_a_field(self):
        results = []
        (
            emit(limit=9, every='1s', start='2016-01-01')
            | (
                put(foo='bar'),
                put(foo='baz')
            ) | reduce(count=count(), every='3s', by=['foo'], reset=False)
            | memory(results)
        ).execute()
        expect(results).to.have.length(6)

        for point in [
                {'time': '2016-01-01T00:00:00.000Z', 'count': 3, 'foo': 'bar'},
                {'time': '2016-01-01T00:00:03.000Z', 'count': 6, 'foo': 'bar'},
                {'time': '2016-01-01T00:00:06.000Z', 'count': 9, 'foo': 'bar'},
                {'time': '2016-01-01T00:00:00.000Z', 'count': 3, 'foo': 'baz'},
                {'time': '2016-01-01T00:00:03.000Z', 'count': 6, 'foo': 'baz'},
                {'time': '2016-01-01T00:00:06.000Z', 'count': 9, 'foo': 'baz'}]:
            expect(results).to.contain(point)

    def test_reduce_with_timeless_points(self):
        results = []
        (
            emit(limit=10, every='1s', start='2016-01-01')
            | remove('time')
            | reduce(count=count())
            | memory(results)
        ).execute()
        expect(results).to.eq([{'count': 10}])

    def test_reduce_monthly_day_counts(self):
        results = []
        (
            emit(limit=365, start='2015-01-01', every='1d')
            | reduce(count=count(), every='1 month')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'time': '2015-01-01T00:00:00.000Z', 'count': 31},
            {'time': '2015-02-01T00:00:00.000Z', 'count': 28},
            {'time': '2015-03-01T00:00:00.000Z', 'count': 31},
            {'time': '2015-04-01T00:00:00.000Z', 'count': 30},
            {'time': '2015-05-01T00:00:00.000Z', 'count': 31},
            {'time': '2015-06-01T00:00:00.000Z', 'count': 30},
            {'time': '2015-07-01T00:00:00.000Z', 'count': 31},
            {'time': '2015-08-01T00:00:00.000Z', 'count': 31},
            {'time': '2015-09-01T00:00:00.000Z', 'count': 30},
            {'time': '2015-10-01T00:00:00.000Z', 'count': 31},
            {'time': '2015-11-01T00:00:00.000Z', 'count': 30},
            {'time': '2015-12-01T00:00:00.000Z', 'count': 31},
        ])

    def test_reduce_yearl_day_counts(self):
        results = []
        (
            emit(limit=365 * 8 + 2, start='2000-01-01', every='1d')
            | reduce(days=count(), every='1 year')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'time': '2000-01-01T00:00:00.000Z', 'days': 366},
            {'time': '2001-01-01T00:00:00.000Z', 'days': 365},
            {'time': '2002-01-01T00:00:00.000Z', 'days': 365},
            {'time': '2003-01-01T00:00:00.000Z', 'days': 365},
            {'time': '2004-01-01T00:00:00.000Z', 'days': 366},
            {'time': '2005-01-01T00:00:00.000Z', 'days': 365},
            {'time': '2006-01-01T00:00:00.000Z', 'days': 365},
            {'time': '2007-01-01T00:00:00.000Z', 'days': 365}
        ])

    def test_reduce_with_single_empty_every_interval(self):
        results = []
        (
            emit(points=[
                {'time': '2016-01-01T00:00:00.000Z', 'value': 1},
                {'time': '2016-01-01T01:00:00.000Z', 'value': 2},
                {'time': '2016-01-01T02:00:00.000Z', 'value': 3},
                {'time': '2016-01-01T04:00:00.000Z', 'value': 4},
                {'time': '2016-01-01T05:00:00.000Z', 'value': 5},
            ])
            | reduce(count=count(), every='1h')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'time': '2016-01-01T00:00:00.000Z', 'count': 1},
            {'time': '2016-01-01T01:00:00.000Z', 'count': 1},
            {'time': '2016-01-01T02:00:00.000Z', 'count': 1},
            {'time': '2016-01-01T03:00:00.000Z', 'count': 0},
            {'time': '2016-01-01T04:00:00.000Z', 'count': 1},
            {'time': '2016-01-01T05:00:00.000Z', 'count': 1},
        ])

    def test_reduce_with_multiple_empty_every_interval(self):
        results = []
        (
            emit(points=[
                {'time': '2016-01-01T00:00:00.000Z'},
                {'time': '2016-01-01T01:00:00.000Z'},
                {'time': '2016-01-01T02:00:00.000Z'},
                {'time': '2016-01-01T05:00:00.000Z'},
                {'time': '2016-01-01T06:00:00.000Z'},
                {'time': '2016-01-01T09:00:00.000Z'},
                {'time': '2016-01-01T10:00:00.000Z'}
            ])
            | reduce(count=count(), every='1h')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'time': '2016-01-01T00:00:00.000Z', 'count': 1},
            {'time': '2016-01-01T01:00:00.000Z', 'count': 1},
            {'time': '2016-01-01T02:00:00.000Z', 'count': 1},
            {'time': '2016-01-01T03:00:00.000Z', 'count': 0},
            {'time': '2016-01-01T04:00:00.000Z', 'count': 0},
            {'time': '2016-01-01T05:00:00.000Z', 'count': 1},
            {'time': '2016-01-01T06:00:00.000Z', 'count': 1},
            {'time': '2016-01-01T07:00:00.000Z', 'count': 0},
            {'time': '2016-01-01T08:00:00.000Z', 'count': 0},
            {'time': '2016-01-01T09:00:00.000Z', 'count': 1},
            {'time': '2016-01-01T10:00:00.000Z', 'count': 1}
        ])

    def test_reduce_by_a_nested_field(self):
        results = []
        (
            emit(points=[
                {'time': '2016-01-01T00:00:00.000Z', 'author': {'name': 'joe'}},
                {'time': '2016-01-01T00:01:00.000Z', 'author': {'name': 'joe'}},
                {'time': '2016-01-01T00:02:00.000Z', 'author': {'name': 'bob'}},
                {'time': '2016-01-01T00:03:00.000Z', 'author': {'name': 'joe'}},
                {'time': '2016-01-01T00:03:00.000Z', 'author': {'name': 'bob'}},
            ])
            | reduce(count=count(), by=['author.name'])
            | memory(results)
        ).execute()
        expect(results).to.contain({
            'time': '2016-01-01T00:00:00.000Z',
            'author': {
                'name': 'bob'
            },
            'count': 2
        })
        expect(results).to.contain({
            'time': '2016-01-01T00:00:00.000Z',
            'author': {
                'name': 'joe'
            },
            'count': 3
        })

    def test_reduce_maximum_on_nested_field(self):
        results = []
        (
            emit(points=[
                {'time': '2016-01-01T00:00:00.000Z', 'nested': {'value': 1}},
                {'time': '2016-01-01T00:01:00.000Z', 'nested': {'value': 5}},
                {'time': '2016-01-01T00:02:00.000Z', 'nested': {'value': 3}},
                {'time': '2016-01-01T00:03:00.000Z', 'nested': {'value': 2}},
                {'time': '2016-01-01T00:03:00.000Z', 'nested': {'value': 7}},
            ])
            | reduce(maximum=maximum('nested.value'))
            | remove('time')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'maximum': 7}
        ])



"""
base streamer unittests
"""

import unittest

from robber import expect

from flume.adapters.streamers import Streamer, register_streamer, get_streamer
from flume.exceptions import FlumineException
from test.unit.util import FakeIO


class BaseStreamersTest(unittest.TestCase):

    def test_register_non_unique_streamer_fails(self):
        register_streamer('non-unique', object())
        try:
            register_streamer('non-unique', object())
            raise Exception('previous statement should have failed')
        except FlumineException as exception:
            expect(exception.message).to.contain('"non-unique" streamer already registered')

    def test_register_and_get_streamer_by_name(self):
        class TestStreamer(Streamer):

            def __init__(self):
                pass

        register_streamer('another', TestStreamer)
        result = get_streamer('another')
        self.assertIsInstance(result, TestStreamer)

    def test_get_streamer_of_inexistent_streamer_fails(self):
        try:
            get_streamer('non-existent')
            raise Exception('previous statement should have failed')
        except FlumineException as exception:
            expect(exception.message).to.contain('"non-existent" streamer not registered')

    def test_new_streamer_must_have_a_read_method(self):
        try:
            class TestStreamer(Streamer):
                def __init__(self):
                    pass

            stream = FakeIO()
            streamer = TestStreamer()
            streamer.read(stream)
            raise Exception('previous statement should have failed')
        except FlumineException as exception:
            expect(exception.message).to.contain('you must implement the read method')

    def test_new_streamer_must_have_a_write_method(self):
        try:
            class TestStreamer(Streamer):
                def __init__(self):
                    pass

            stream = FakeIO()
            streamer = TestStreamer()
            streamer.write(stream, [])
            raise Exception('previous statement should have failed')
        except FlumineException as exception:
            expect(exception.message).to.contain('you must implement the write method')



"""
processor init module
"""
from flume.procs.keep import keep
from flume.procs.remove import remove
from flume.procs.put import put
from flume.procs.head import head
from flume.procs.tail import tail

# pylint: disable=redefined-builtin
from flume.procs.filter import filter

# pylint: disable=redefined-builtin
from flume.procs.reduce import reduce

from flume.procs.seq import seq

from flume.procs.reorder import reorder
from flume.procs.sort import sort

# set procs
from flume.procs.diff import diff
from flume.procs.intersect import intersect
from flume.procs.union import union



"""
seq proc
"""
from flume import node, queue

class pipe(node):

    def loop(self):
        while self.running:
            points = self.pull()
            self.push(points)

    def push(self, points):
        node.push(self, [point for point in points if point != self.EOF])

class seq(node):

    name = 'seq'

    def __init__(self, *pipelines):
        node.__init__(self)
        self.pipelines = pipelines

    def loop(self):
        for pipeline in self.pipelines:
            outputs = list(self.outputs)

            node.init_node(pipeline,
                           inputs=None,
                           outputs=[queue()],
                           parent=None,
                           source=None)

            connector = pipe()
            node.init_node(connector,
                           inputs=pipeline.outputs,
                           outputs=outputs,
                           parent=pipeline,
                           source=None)

            connector.execute(wait=True)



"""
reorder processor
"""
import bisect

from flume import logger, moment, node


class reorder(node):

    name = 'reorder'

    def __init__(self, delay=None):
        node.__init__(self)
        self.delay = delay

    def loop(self):
        points = []
        btimes = []
        buffered = []
        last = None

        while self.running:
            # reorder can't get stuck on a single input because it doesn't
            # have points if its going to get points in realtime and emit them
            # so we set the timeout to 1ms and that way we'll be able to process
            # a set of live inputs and never delay them by more than the `delay`
            # specified
            points = self.pull(wait=True, timeout=0.001)

            for point in points:
                # each point is timestamped upon arrival and only emitted
                # after the `delay` specified
                point.__arrival_time__ = moment.now()

                if 'time' not in point.keys():
                    buffered.append(point)

                else:
                    index = bisect.bisect_left(btimes, point.time)
                    btimes.insert(index, point.time)
                    buffered.insert(index, point)

            result = []
            keep = []
            for point in buffered:
                if 'time' not in point:
                    del point.__arrival_time__
                    result.append(point)

                elif (point.__arrival_time__ + self.delay) <= moment.now():
                    if last is not None and last.time > point.time:
                        logger.warn('dropping point %s, arrived after %s' %
                                    (point, last))
                        continue
                
                    del point.__arrival_time__
                    result.append(point)
                    last = point

                else:
                    keep.append(point)

            buffered = keep
            btimes = [point.time for point in buffered]

            if result != []:
                self.push(result)

        result = []
        for point in buffered:
            del point.__arrival_time__
            result.append(point)

        if result != []:
            self.push(result)



"""
setup.py
"""
from setuptools import setup, find_packages

def load(filename):
    with open(filename) as input_file:
        return input_file.read().split('\n'),

setup(
    name='flume',
    version=open('flume/VERSION').read(),
    author='Rodney Gomes',
    author_email='rodneygomes@gmail.com',
    url='',
    install_requires=load('requirements.txt'),
    tests_require=load('test-requirements.txt'),
    test_suite='test.unit',
    keywords=[''],
    py_modules=['flume'],
    packages=find_packages(exclude=['test']),
    include_package_data=True,

    entry_points={
        'console_scripts': [
            'flume=flume.cli:main'
        ]
    },

    license='Apache 2.0 License',
    description='flume (pronounced: floom) is a stream processing framework',
    long_description='README: http://github.com/rlgomes/flume',
)



"""
diff proc
"""
from collections import OrderedDict

from flume import node


class diff(node):

    name = 'diff'

    def __init__(self, *fieldnames):
        node.__init__(self)
        self.fieldnames = fieldnames

    def loop(self):
        inputs_count = self.input_count()

        if len(self.fieldnames) == 0:
            previous_point = None

            while self.running:
                points = self.pull()

                for point in points:
                    flume_path = point.__meta__.flume_path

                    if previous_point is None:
                        previous_point = point
                        previous_point.__meta__.inputs = set([flume_path])

                    elif previous_point.time == point.time:
                        previous_point.__meta__.inputs.add(flume_path)
                        previous_point.update(point)

                    else:
                        # only push the points that didn't appear in all inputs
                        if inputs_count == 1 or \
                           len(previous_point.__meta__.inputs) != inputs_count:
                            self.push(previous_point)

                        previous_point = point
                        previous_point.__meta__.inputs = set([flume_path])

            # only push the points that didn't appear in all inputs
            if inputs_count == 1 or len(previous_point.__meta__.inputs) != inputs_count:
                self.push(previous_point)

        else:
            # when diffing on non `time` field we basically buffer every point
            # and emit when we have a complete window

            # XXX: non time based result should have a limit on maximum
            #      amount of points to handle like 100K so we don't run out of
            #      memory

            result = OrderedDict()
            fieldnames = self.fieldnames
            while self.running:
                points = self.pull()

                for point in points:
                    flume_path = point.__meta__.flume_path
                    key = '.'.join([str(point[fname]) for fname in fieldnames])

                    if key not in result:
                        result[key] = point
                        result[key].__meta__.inputs = set([flume_path])

                    else:
                        result[key].__meta__.inputs.add(flume_path)
                        result[key].update(point)

            points = []

            for point in result.values():
                # only push the points that didn't appear in all inputs
                if len(point.__meta__.inputs) != inputs_count:
                    points.append(point)

            self.push(points)



"""
streamers init module
"""
from flume.adapters.streamers.base import Streamer
from flume.adapters.streamers.csver import CSV
from flume.adapters.streamers.jsonler import JSONL
from flume.adapters.streamers.jsoner import JSON
from flume.adapters.streamers.grok import Grok
from flume.exceptions import FlumineException

_STREAMERS = {}


def get_streamer(name, *args, **kwargs):
    if name in _STREAMERS:
        return _STREAMERS[name](*args, **kwargs)

    else:
        raise FlumineException('"%s" streamer not registered' % name)

def register_streamer(name, streamer):
    if name in _STREAMERS.keys():
        raise FlumineException('"%s" streamer already registered' % name)

    _STREAMERS[name] = streamer

register_streamer('jsonl', JSONL)
register_streamer('json', JSON)
register_streamer('csv', CSV)
register_streamer('grok', Grok)



"""
"""
from flume.exceptions import FlumineException

_VIEWS = {}

def get_view(name, **kwargs):
    if name not in _VIEWS:
        raise FlumineException('view with name "%s", not found' % name)

    return _VIEWS[name](**kwargs)

def register_view(name, view):
    if name in _VIEWS:
        raise FlumineException('view with name "%s", already registered' % name)

    _VIEWS[name] = view

from flume.sinks.views.pygalview import PyGal
from flume.sinks.views.gnuplot import Gnuplot

register_view('pygal', PyGal)
register_view('gnuplot', Gnuplot)



# coding utf-8
"""
http adapter unittests
"""

import json
import sys
import tempfile
import unittest
import zlib

import mock

from robber import expect

from flume import *
from test.unit.util import FakeIO


class StdioTest(unittest.TestCase):

    @classmethod
    def tearDownClass(cls):
        # make sure to set the stdin/stdout hooks in stdio back to the
        # right values
        stdio.stdin = sys.stdin
        stdio.stdout = sys.stdout

    def test_stdio_can_read_from_empty(self):
        stdio.stdin = FakeIO('')
        results = []

        (
            read('stdio')
            | memory(results)
        ).execute()
        expect(results).to.eq([])

    def test_stdio_can_read_and_strip_ansi_sequences_from_terminal(self):
        stdio.stdin = FakeIO('''\033[1;32m pid cmd cpu\033[0m
1 init 0.0
2 bash 2.0
25 blah 3.0''')
        results = []

        (
            read('stdio', format='csv', delimiter=' ', strip_ansi=True)
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'pid': '1', 'cmd': 'init', 'cpu': '0.0'},
            {'pid': '2', 'cmd': 'bash', 'cpu': '2.0'},
            {'pid': '25', 'cmd': 'blah', 'cpu': '3.0'}
        ])

    def test_stdio_can_read_and_strip_ansi_sequences_from_jsonl(self):
        stdio.stdin = FakeIO('{"time": "2016-01-01T00:00:00.000Z", "foo": "\033[1;32mbar\033[0m"}')
        results = []

        (
            read('stdio', format='jsonl', strip_ansi=True)
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {"time": "2016-01-01T00:00:00.000Z", "foo": "bar"}
        ])

    def test_stdio_can_read_and_strip_ansi_sequences_from_json(self):
        stdio.stdin = FakeIO('[{"time": "2016-01-01T00:00:00.000Z", "foo": "\033[1;32mbar\033[0m"}]')
        results = []

        (
            read('stdio', format='json', strip_ansi=True)
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {"time": "2016-01-01T00:00:00.000Z", "foo": "bar"}
        ])

    def test_stdio_can_read_and_not_strip_ansi_sequences_from_json(self):
        stdio.stdin = FakeIO('[{"time": "2016-01-01T00:00:00.000Z", "foo": "\033[1;32mbar\033[0m"}]')
        results = []

        (
            read('stdio', format='json')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {"time": "2016-01-01T00:00:00.000Z", "foo": "\033[1;32mbar\033[0m"}
        ])

    def test_stdio_can_read_a_single_point(self):
        stdio.stdin = FakeIO('{"time": "2016-01-01T00:00:00.000Z", "foo": "bar"}')
        results = []

        (
            read('stdio', format='jsonl')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {"time": "2016-01-01T00:00:00.000Z", "foo": "bar"}
        ])

    def test_stdio_can_read_multiple_points(self):
        stdio.stdin = FakeIO('{"time": "2016-01-01T00:00:00.000Z", "foo": 1}\n' +
                               '{"time": "2016-01-01T00:01:00.000Z", "foo": 2}\n' +
                               '{"time": "2016-01-01T00:02:00.000Z", "foo": 3}\n')

        results = []

        (
            read('stdio', format='jsonl')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {"time": "2016-01-01T00:00:00.000Z", "foo": 1},
            {"time": "2016-01-01T00:01:00.000Z", "foo": 2},
            {"time": "2016-01-01T00:02:00.000Z", "foo": 3}
        ])

    def test_stdio_can_read_a_single_timeless_point(self):
        stdio.stdin = FakeIO('{"foo": "bar"}')
        results = []

        (
            read('stdio', format='jsonl')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {"foo": "bar"}
        ])

    def test_stdio_can_read_multiple_timeless_points(self):
        stdio.stdin = FakeIO('{"foo": 1}\n{"foo": 2}\n{"foo": 3}\n')
        results = []

        (
            read('stdio', format='jsonl')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {"foo": 1},
            {"foo": 2},
            {"foo": 3}
        ])

    def test_stdio_can_read_multiple_points_with_custom_timefield(self):
        stdio.stdin = FakeIO('{"bar": "2016-01-01T00:00:00.000Z", "foo": 1}\n' +
                               '{"bar": "2016-01-01T00:01:00.000Z", "foo": 2}\n' +
                               '{"bar": "2016-01-01T00:02:00.000Z", "foo": 3}\n')

        results = []

        (
            read('stdio', format='jsonl', time='bar')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {"time": "2016-01-01T00:00:00.000Z", "foo": 1},
            {"time": "2016-01-01T00:01:00.000Z", "foo": 2},
            {"time": "2016-01-01T00:02:00.000Z", "foo": 3}
        ])

    def test_stdio_can_write_nothing(self):
        stdio.stdin = FakeIO('')
        stdio.stdout = FakeIO('')
        (
            read('stdio')
            | write('stdio')
        ).execute()
        expect(stdio.stdout.getvalue()).to.eq('')

    def test_stdio_can_write_a_single_point(self):
        stdio.stdout = FakeIO('')
        (
            emit(limit=1, start='2016-01-01')
            | write('stdio')
        ).execute()
        expect(json.loads(stdio.stdout.getvalue())).to.eq({
            'time': '2016-01-01T00:00:00.000Z'
        })

    def test_stdio_can_write_a_multiple_points(self):
        stdout = FakeIO('')
        stdio.stdout = stdout
        (
            emit(limit=5, start='2016-01-01')
            | put(count=count())
            | write('stdio')
        ).execute()

        results = []
        stdio.stdin = FakeIO(stdout.getvalue())
        (
            read('stdio')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'time': '2016-01-01T00:00:00.000Z', 'count': 1},
            {'time': '2016-01-01T00:00:01.000Z', 'count': 2},
            {'time': '2016-01-01T00:00:02.000Z', 'count': 3},
            {'time': '2016-01-01T00:00:03.000Z', 'count': 4},
            {'time': '2016-01-01T00:00:04.000Z', 'count': 5}
        ])

    def test_stdio_can_write_a_single_timeless_point(self):
        stdio.stdout = FakeIO('')
        (
            emit(limit=1, start='2016-01-01')
            | put(foo='bar')
            | remove('time')
            | write('stdio')
        ).execute()
        expect(json.loads(stdio.stdout.getvalue())).to.eq({
            'foo': 'bar'
        })

    def test_stdio_can_write_a_multiple_timeless_points(self):
        stdout = FakeIO('')
        stdio.stdout = stdout
        (
            emit(limit=5, start='2016-01-01')
            | put(count=count())
            | remove('time')
            | write('stdio')
        ).execute()

        results = []
        stdio.stdin = FakeIO(stdout.getvalue())
        (
            read('stdio')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'count': 1},
            {'count': 2},
            {'count': 3},
            {'count': 4},
            {'count': 5}
        ])

    def test_stdio_can_read_a_single_point_from_a_file(self):
        _, tmpfile = tempfile.mkstemp()

        with open(tmpfile, 'w') as output:
            output.write('{"time": "2016-01-01T00:00:00.000Z", "foo": "bar"}')

        results = []
        (
            read('stdio', file=tmpfile)
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {"time": "2016-01-01T00:00:00.000Z", "foo": "bar"}
        ])

    def test_stdio_can_write_a_single_point_to_a_file(self):
        _, tmpfile = tempfile.mkstemp()

        (
            emit(limit=1, start='2016-01-01')
            | write('stdio', file=tmpfile)
        ).execute()

        with open(tmpfile, 'r') as tmp_input:
            expect(json.loads(tmp_input.read())).to.eq({
                'time': '2016-01-01T00:00:00.000Z'
            })


    def test_stdio_can_write_override_single_point_to_a_file(self):
        _, tmpfile = tempfile.mkstemp()

        (
            emit(limit=1, start='2015-01-01')
            | write('stdio', file=tmpfile)
        ).execute()

        (
            emit(limit=1, start='2016-01-01')
            | write('stdio', file=tmpfile, append=False)
        ).execute()

        with open(tmpfile, 'r') as tmp_input:
            expect(json.loads(tmp_input.read())).to.eq({
                'time': '2016-01-01T00:00:00.000Z'
            })

    def test_stdio_can_write_append_multiple_points_to_a_file(self):
        _, tmpfile = tempfile.mkstemp()

        (
            emit(limit=1, start='2015-01-01')
            | write('stdio', file=tmpfile)
        ).execute()

        (
            emit(limit=1, start='2016-01-01')
            | write('stdio', file=tmpfile, append=True)
        ).execute()
        
        results = []

        (
            read('stdio', file=tmpfile)
            | memory(results)
        ).execute() 
        expect(results).to.eq([
            {'time': '2015-01-01T00:00:00.000Z'},
            {'time': '2016-01-01T00:00:00.000Z'}
        ])

    def test_stdio_raises_exception_on_invalid_compression_on_write(self):
        with self.assertRaisesRegexp(FlumineException, 'unsupported compression \[bananas\]'):
            (
                emit(limit=1, start='2016-01-01')
                | write('stdio', format='jsonl', compression='bananas')
            ).execute()

    def test_stdio_raises_exception_on_invalid_compression_on_read(self):
        with self.assertRaisesRegexp(FlumineException, 'unsupported compression \[bananas\]'):
            (
                read('stdio', format='jsonl', compression='bananas')
            ).execute()

    def test_stdio_can_write_and_read_a_single_point_in_jsonl_with_zlib_compression(self):
        _, tmpfile = tempfile.mkstemp()
        (
            emit(limit=1, start='2016-01-01')
            | write('stdio', file=tmpfile, format='jsonl', compression='zlib')
        ).execute()

        results = []
        (
            read('stdio', file=tmpfile, format='jsonl', compression='zlib')
            | memory(results)
        ).execute()
        expect(results).to.eq([{"time": "2016-01-01T00:00:00.000Z"}])

    def test_stdio_can_write_and_read_a_single_point_in_jsonl_with_gzip_compression(self):
        _, tmpfile = tempfile.mkstemp()
        (
            emit(limit=1, start='2016-01-01')
            | write('stdio', file=tmpfile, format='jsonl', compression='gzip')
        ).execute()

        results = []
        (
            read('stdio', file=tmpfile, format='jsonl', compression='gzip')
            | memory(results)
        ).execute()
        expect(results).to.eq([{"time": "2016-01-01T00:00:00.000Z"}])

    def test_stdio_can_write_and_read_a_single_point_in_csv_with_deflate_compression(self):
        _, tmpfile = tempfile.mkstemp()
        (
            emit(limit=1, start='2016-01-01')
            | write('stdio', file=tmpfile, format='csv', compression='deflate')
        ).execute()

        results = []
        (
            read('stdio', file=tmpfile, format='csv', compression='deflate')
            | memory(results)
        ).execute()
        expect(results).to.eq([{"time": "2016-01-01T00:00:00.000Z"}])



"""
read source
"""

from flume import adapters, logger, moment, node


class read(node):
    """
    read

    usage:
        read('adapter', [adapter arguments]) | ...  
    The read source can be used to read points from various adapters of which
    include the following:
    """

    name = 'read'

    def __init__(self,
                 adapter,
                 time='time',
                 **kwargs):
        node.__init__(self)
        cls = adapters.get_adapter(adapter, 'read')

        self.instance = cls(**kwargs)
        self.read = self.instance.read
        self.time = time

    def loop(self):
        for points in self.read():

            for point in points:
                if self.time in point:
                    point.time = moment.date(point[self.time])
                    if self.time != 'time':
                        del point[self.time]

            self.push(points)




import unittest

from flume import *
from flume.adapters.adapter import adapter

from robber import expect


class dummy(adapter):

    name = 'dummy'

    def __init__(self, result=[]):
        self.result = result

    def read(self):
        return self.result

    def write(self, points):
        for point in points:
            self.result.append(point.json())

register_adapter(dummy)

class WriteTest(unittest.TestCase):

    def test_write_can_handle_an_empty_stream(self):
        result = []
        (
            emit(points=[])
            | write('dummy', result=result)
        ).execute()
        expect(result).to.eq([])

    def test_write_can_handle_a_few_points(self):
        result = []
        (
            emit(points=[
                {'foo': 1},
                {'foo': 1},
                {'foo': 1}
            ])
            | write('dummy', result=result)
        ).execute()
        expect(result).to.eq([
            {'foo': 1},
            {'foo': 1},
            {'foo': 1}
        ])



"""
adapters base module
"""

from flume.adapters.adapter import adapter
from flume.adapters.elastic import elastic
from flume.adapters.http import http
from flume.adapters.stdio import stdio
from flume.exceptions import FlumineException


_ADAPTERS = {}


def register_adapter(cls):
    """
    register a new adapter class
    """
    if cls.name in _ADAPTERS.keys():
        raise FlumineException('"%s" adapter already registered' % cls.name)

    _ADAPTERS[cls.name] = cls

def get_adapter(name, operation):
    """
    get a previously registered adapter class by name
    """
    if name not in _ADAPTERS.keys():
        raise FlumineException('"%s" adapter not registered' % name)

    result = _ADAPTERS[name]

    if getattr(result, operation) == getattr(adapter, operation):
        raise FlumineException('"%s" adapter does not support %s' %
                               (name, operation))

    return result



from flume import reducer

class maximum(reducer):

    def __init__(self, fieldname):
        self.fieldname = fieldname
        self.max = None

    def update(self, point):
        if self.max:
            if point[self.fieldname] > self.max:
                self.max = point[self.fieldname]
        else:
            self.max = point[self.fieldname]

    def result(self):
        return self.max

    def reset(self):
        self.max = None
